@@ -1,10 +1,9 @@
 /*
- * Copyright 1996-2003 by Hans Reiser, licensing governed by 
+ * Copyright 1996-2004 by Hans Reiser, licensing governed by 
  * reiserfsprogs/README 
  */
 
 #include "fsck.h"
-#include <stdlib.h>
 
 reiserfs_bitmap_t * bad_unfm_in_tree_once_bitmap;
 
@@ -28,10 +27,10 @@ struct buffer_head * make_buffer (int dev, unsigned long blocknr, int size, char
 
 
 int find_not_of_one_file(struct key * to_find, struct key * key) {
-    if ((get_key_objectid (to_find) != ~0ul) &&
+    if ((get_key_objectid (to_find) != ~(__u32)0) &&
         (get_key_objectid (to_find) != get_key_objectid (key)))
         return 1;
-    if ((get_key_dirid (to_find) != ~0ul) &&
+    if ((get_key_dirid (to_find) != ~(__u32)0) &&
         (get_key_dirid (to_find) != get_key_dirid (key)))
         return 1;
     return 0;
@@ -106,7 +105,7 @@ static void indirect_in_tree (struct buffer_head * bh,
     unp = (__u32 *)B_I_PITEM (bh, ih);
     
     for (i = 0; i < I_UNFM_NUM (ih); i ++) {
-	unfm_ptr = le32_to_cpu (unp[i]);
+	unfm_ptr = d32_get (unp, i);
 	if (unfm_ptr == 0)
 	    continue;
 	if ((ret = still_bad_unfm_ptr_1 (unfm_ptr)))
@@ -246,7 +245,8 @@ static void get_max_buffer_key (struct buffer_head * bh, struct key * key)
 
 int tree_is_empty (void)
 {
-    return (get_sb_root_block (fs->fs_ondisk_sb) == ~0ul || get_sb_root_block (fs->fs_ondisk_sb) == 0) ? 1 : 0;
+    return (get_sb_root_block (fs->fs_ondisk_sb) == ~(__u32)0 || 
+	    get_sb_root_block (fs->fs_ondisk_sb) == 0) ? 1 : 0;
 }
 
 
@@ -394,8 +394,8 @@ static void pass1_correct_leaf (reiserfs_filsys_t * fs,
 	/* correct indirect items */
 	ind_item = (__u32 *)B_I_PITEM (bh, ih);
 
-	for (j = 0; j < I_UNFM_NUM (ih); j ++, ind_item ++) {
-	    unfm_ptr = le32_to_cpu (*ind_item);
+	for (j = 0; j < I_UNFM_NUM (ih); j ++) {
+	    unfm_ptr = d32_get (ind_item, j);
 
 	    if (!unfm_ptr)
 		continue;
@@ -409,7 +409,7 @@ static void pass1_correct_leaf (reiserfs_filsys_t * fs,
 	    /* 1. zero slots pointing to a leaf */
 	    if (is_used_leaf (unfm_ptr)) {
 		dirty ++;
-		*ind_item = 0;
+		d32_put (ind_item, j, 0);
 		pass_1_stat (fs)->pointed_leaves ++;
 		continue;
 	    }
@@ -424,7 +424,7 @@ static void pass1_correct_leaf (reiserfs_filsys_t * fs,
 		else {
 		    /* Yes, we have seen this pointer already, zero other pointers to it. */
 		    dirty ++;
-		    *ind_item = 0;
+		    d32_put (ind_item, j, 0);
 		    pass_1_stat (fs)->non_unique_pointers ++;
 		    continue;
 		}
@@ -543,8 +543,9 @@ static void find_allocable_blocks (reiserfs_filsys_t * fs)
 	}
 
 	/* make allocable not leaves, not bad blocks */
-	if (!is_used_leaf (i) &&
-	    (!fs->fs_badblocks_bm || !reiserfs_bitmap_test_bit (fs->fs_badblocks_bm, i))) {
+	if (!is_used_leaf (i) && (!fs->fs_badblocks_bm || 
+				  !reiserfs_bitmap_test_bit (fs->fs_badblocks_bm, i))) 
+	{
 	    /* this is not leaf and it is not pointed by found indirect items,
                so it does not contains anything valuable */
 	    make_allocable (i);
